import logging

from typing import Any, Dict, List

import psycopg2.extensions

from django.db import connection, transaction
from django.db.models import QuerySet
from rapidsms.models import Backend

from smpp_gateway.models import MOMessage, MTMessage

logger = logging.getLogger(__name__)


def pg_listen(channel: str) -> psycopg2.extensions.connection:
    """Return a new connection listening for notifications on `channel`.

    Can be used like:
    >>> conn = pg_listen("my_channel")
    >>> conn.poll()
    >>> for notify in conn.notifies:
    >>>    print(f"I got a notification: {notify}")
    """
    with connection.cursor() as cursor:
        pg_conn = connection.connection
        pg_conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        # https://gist.github.com/pkese/2790749
        cursor.execute(f"LISTEN {channel};")
        logger.info(f"Waiting for notifications on channel '{channel}'")

    return pg_conn


def pg_notify(channel: str):
    """Send a notification on `channel` with empty payload."""
    with connection.cursor() as cursor:
        cursor.execute(f"NOTIFY {channel};")


def get_mt_messages_to_send(limit: int, backend: Backend) -> List[Dict[str, Any]]:
    """Fetches up to `limit` messages intended for `backend`, updates their
    status to SENDING, and returns select fields from the model.
    """
    with transaction.atomic():
        smses = list(
            MTMessage.objects.filter(status=MTMessage.Status.NEW, backend=backend)
            .select_for_update(skip_locked=True)
            .values("id", "short_message", "params")[:limit]
        )
        if smses:
            pks = [sms["id"] for sms in smses]
            logger.debug(
                f"get_mt_messages_to_send: Marking {pks} as {MTMessage.Status.SENDING.label}"
            )
            MTMessage.objects.filter(pk__in=pks).update(status=MTMessage.Status.SENDING)
    return smses


def get_mo_messages_to_process(limit: int = 1) -> QuerySet[MOMessage]:
    """Fetches up to `limit` incoming messages while updating their
    status to PROCESSING.
    """
    with transaction.atomic():
        smses = (
            MOMessage.objects.filter(status=MOMessage.Status.NEW)
            .select_related("backend")
            .select_for_update(skip_locked=True, of=("self",))[:limit]
        )
        pks = [sms.pk for sms in smses]
        logger.debug(
            f"get_mo_messages_to_process: Marking {pks} as {MOMessage.Status.PROCESSING.label}"
        )
        MOMessage.objects.filter(pk__in=pks).update(status=MOMessage.Status.PROCESSING)
    return smses



# Generated by Django 3.2.13 on 2022-05-24 14:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("smpp_gateway", "0002_mtmessagestatus"),
    ]

    operations = [
        migrations.AlterField(
            model_name="momessage",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="params",
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="params",
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="delivery_report",
            field=models.BinaryField(
                help_text="The delivery_report we receive from the MC (via its deliver_sm).",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="message_id",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The message_id we receive from the MC (via its submit_sm_resp).",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="sequence_number",
            field=models.IntegerField(
                help_text="The initial sequence_number established by us when the message was sent."
            ),
        ),
    ]



import factory
import smpplib.consts

from django.utils.timezone import now
from factory.django import DjangoModelFactory
from faker import Faker
from rapidsms.models import Backend

from smpp_gateway.models import MOMessage, MTMessage, MTMessageStatus


class BackendFactory(DjangoModelFactory):
    class Meta:
        model = Backend

    name = factory.Sequence(lambda n: f"test_backend_{n}")


class MOMessageFactory(DjangoModelFactory):
    class Meta:
        model = MOMessage

    create_time = factory.LazyFunction(now)
    modify_time = factory.LazyFunction(now)
    backend = factory.SubFactory(BackendFactory)
    short_message = factory.LazyFunction(
        lambda: Faker().text(max_nb_chars=256).encode("ascii")
    )
    params = factory.Dict(
        {
            "destination_addr": "+46166371877",
            "source_addr": "+46166371876",
        }
    )
    status = MOMessage.Status.NEW


class MTMessageFactory(DjangoModelFactory):
    class Meta:
        model = MTMessage

    create_time = factory.LazyFunction(now)
    modify_time = factory.LazyFunction(now)
    backend = factory.SubFactory(BackendFactory)
    short_message = factory.Faker("sentence")
    params = {}
    status = MTMessage.Status.NEW


class MTMessageStatusFactory(DjangoModelFactory):
    class Meta:
        model = MTMessageStatus

    create_time = factory.LazyFunction(now)
    modify_time = factory.LazyFunction(now)
    mt_message = factory.SubFactory(MTMessageFactory)
    backend = factory.LazyAttribute(lambda o: o.mt_message.backend)
    sequence_number = factory.Sequence(lambda n: n + 1)
    command_status = smpplib.consts.SMPP_ESME_ROK
    message_id = ""
    delivery_report = b""



import logging

from django.db import connection
from django.utils import timezone
from rapidsms.backends.base import BackendBase

from smpp_gateway.models import MTMessage
from smpp_gateway.utils import grouper

logger = logging.getLogger(__name__)


class SMPPGatewayBackend(BackendBase):
    """Outgoing SMS backend for smpp_gateway."""

    # Optional additional params from:
    # https://github.com/python-smpplib/python-smpplib/blob/d9d91beb2d7f37915b13a064bb93f907379342ec/smpplib/command.py#L652-L700
    OPTIONAL_PARAMS = ("source_addr",)

    def configure(self, **kwargs):
        self.send_group_size = kwargs.get("send_group_size", 100)
        self.socket_timeout = kwargs.get("socket_timeout", 5)

    def prepare_request(self, id_, text, identities, context):
        for identity in identities:
            now = timezone.now()
            params = {
                param: context[param]
                for param in self.OPTIONAL_PARAMS
                if param in context
            }
            params["destination_addr"] = identity
            yield {
                "create_time": now,
                "modify_time": now,
                "backend": self.model,
                "short_message": text,
                "params": params,
                "status": MTMessage.Status.NEW,
            }

    def send(self, id_, text, identities, context=None):
        logger.debug("Sending message: %s", text)
        context = context or {}
        kwargs_generator = self.prepare_request(id_, text, identities, context)
        for kwargs_group in grouper(kwargs_generator, self.send_group_size):
            MTMessage.objects.bulk_create(
                [MTMessage(**kwargs) for kwargs in kwargs_group]
            )
            with connection.cursor() as curs:
                curs.execute(f"NOTIFY {self.model.name}")



from multiprocessing.pool import ThreadPool

import pytest

from smpp_gateway.models import MOMessage, MTMessage
from smpp_gateway.queries import (
    get_mo_messages_to_process,
    get_mt_messages_to_send,
    pg_listen,
    pg_notify,
)
from tests.factories import BackendFactory, MOMessageFactory, MTMessageFactory


@pytest.mark.django_db
class TestGetMessagesToSend(object):
    def test_empty(self):
        """Return an empty list if there are no queued outbound messages."""
        messages = get_mt_messages_to_send(100, BackendFactory())
        assert messages == []

    def test_partial_page(self):
        """Return fewer than `limit` messages if appropriate."""
        backend = BackendFactory()
        MTMessageFactory.create_batch(5, backend=backend)

        messages = get_mt_messages_to_send(100, backend)

        assert len(messages) == 5

    def test_full_page(self):
        """Return only `limit` messages if more are present."""
        backend = BackendFactory()
        MTMessageFactory.create_batch(5, backend=backend)

        messages = get_mt_messages_to_send(3, backend)

        assert len(messages) == 3

    def test_new_messages_only(self):
        """Return only messages with status NEW."""
        backend = BackendFactory()
        new_messages = [MTMessageFactory(backend=backend, status=MTMessage.Status.NEW)]
        for other_status in [
            val for val in MTMessage.Status.values if val != MTMessage.Status.NEW
        ]:
            MTMessageFactory.create_batch(
                3,
                backend=backend,
                status=other_status,
            )
        new_messages.append(
            MTMessageFactory(backend=backend, status=MTMessage.Status.NEW)
        )

        messages = get_mt_messages_to_send(10, backend)

        assert len(messages) == 2
        assert {msg["id"] for msg in messages} == {msg.id for msg in new_messages}

    def test_filter_by_backend(self):
        """Return only messages associated with the provided backend."""
        backend_1 = BackendFactory()
        backend_2 = BackendFactory()
        backend_1_messages = MTMessageFactory.create_batch(3, backend=backend_1)
        MTMessageFactory.create_batch(3, backend=backend_2)

        messages = get_mt_messages_to_send(10, backend_1)

        assert len(messages) == 3
        assert {msg["id"] for msg in messages} == {msg.id for msg in backend_1_messages}

    def test_updates_status_to_sending(self):
        """Returned messages should be updated to status SENDING."""
        backend = BackendFactory()
        MTMessageFactory.create_batch(5, backend=backend)

        messages = get_mt_messages_to_send(3, backend)
        returned_message_ids = {msg["id"] for msg in messages}

        for message in MTMessage.objects.all():
            if message.id in returned_message_ids:
                assert message.status == MTMessage.Status.SENDING
            else:
                assert message.status == MTMessage.Status.NEW


@pytest.mark.django_db
class TestGetMessagesToProcess(object):
    def test_empty(self):
        """Return an empty list if there are no queued inbound messages."""
        messages = get_mo_messages_to_process(limit=1)
        assert messages.count() == 0

    def test_partial_page(self):
        """Return fewer than `limit` messages if appropriate."""
        MOMessageFactory.create_batch(5)

        messages = get_mo_messages_to_process(limit=100)

        assert messages.count() == 5

    def test_full_page(self):
        """Return only `limit` messages if more are present."""
        MOMessageFactory.create_batch(5)

        messages = get_mo_messages_to_process(limit=3)

        assert messages.count() == 3

    def test_new_messages_only(self):
        """Return only messages with status NEW."""
        backend = BackendFactory()
        new_messages = [MOMessageFactory(backend=backend, status=MOMessage.Status.NEW)]
        for other_status in [
            val for val in MOMessage.Status.values if val != MOMessage.Status.NEW
        ]:
            MOMessageFactory.create_batch(
                3,
                backend=backend,
                status=other_status,
            )
        new_messages.append(
            MOMessageFactory(backend=backend, status=MOMessage.Status.NEW)
        )

        messages = get_mo_messages_to_process(limit=10)

        assert len(messages) == 2
        assert set(messages) == set(new_messages)

    def test_updates_status_to_processing(self):
        """Returned messages should be updated to status PROCESSING."""
        backend = BackendFactory()
        MOMessageFactory.create_batch(5, backend=backend)

        messages = get_mo_messages_to_process(limit=3)
        returned_message_pks = {msg.pk for msg in messages}

        for message in MOMessage.objects.all():
            if message.pk in returned_message_pks:
                assert message.status == MOMessage.Status.PROCESSING
            else:
                assert message.status == MOMessage.Status.NEW


@pytest.mark.django_db(transaction=True)
class TestConcurrency(object):
    def test_get_mo_messages_concurrently(self):
        """
        Multiple concurrent calls to get_mo_messages_to_process()
        fetch a given message only once.
        """
        message_count = 50
        backend = BackendFactory()
        new_messages = MOMessageFactory.create_batch(
            message_count,
            backend=backend,
            status=MOMessage.Status.NEW,
        )

        def get_one_message(_):
            return get_mo_messages_to_process(limit=1)[0]

        pool = ThreadPool(processes=10)
        messages = pool.map(get_one_message, range(message_count))

        assert len(messages) == message_count
        assert set(messages) == set(new_messages)

    def test_get_mt_messages_concurrently(self):
        """
        Multiple concurrent calls to get_mt_messages_to_send()
        fetch a given message only once.
        """
        message_count = 50
        backend = BackendFactory()
        new_messages = MTMessageFactory.create_batch(
            message_count,
            backend=backend,
            status=MTMessage.Status.NEW,
        )

        def get_one_message(_):
            return get_mt_messages_to_send(limit=1, backend=backend)[0]

        pool = ThreadPool(processes=10)
        messages = pool.map(get_one_message, range(message_count))

        assert len(messages) == message_count
        assert {msg["id"] for msg in messages} == {msg.id for msg in new_messages}


@pytest.mark.django_db(transaction=True)
class TestNotifications(object):
    def test_listen_notify_new_messages(self):
        """`pg_notify` should publish messages that can be read by the
        connection made from `pg_listen`.
        """
        listen_conn = pg_listen("test_channel")

        listen_conn.poll()
        assert len(listen_conn.notifies) == 0

        for _ in range(5):
            pg_notify("test_channel")

        listen_conn.poll()
        assert len(listen_conn.notifies) == 5

    def test_listen_notify_preexisting_messages(self):
        """The connection made from `pg_listen` cannot recieve messages sent
        before it was created.
        """
        for _ in range(4):
            pg_notify("test_channel")

        listen_conn = pg_listen("test_channel")
        listen_conn.poll()
        assert len(listen_conn.notifies) == 0



# Generated by Django 3.2.16 on 2022-12-06 21:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("smpp_gateway", "0004_translations"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="mtmessagestatus",
            options={
                "verbose_name": "mobile-terminated message status",
                "verbose_name_plural": "mobile-terminated message statuses",
            },
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="command_status",
            field=models.IntegerField(
                choices=[
                    (0, "No Error"),
                    (1, "Message Length is invalid"),
                    (2, "Command Length is invalid"),
                    (3, "Invalid Command ID"),
                    (4, "Incorrect BIND Status for given command"),
                    (5, "ESME Already in Bound State"),
                    (6, "Invalid Priority Flag"),
                    (7, "<Desc Not Set>"),
                    (8, "System Error"),
                    (10, "Invalid Source Address"),
                    (11, "Invalid Destination Address"),
                    (12, "Invalid Message ID"),
                    (13, "Bind Failed"),
                    (14, "Invalid Password"),
                    (15, "Invalid System ID"),
                    (17, "Cancel SM Failed"),
                    (19, "Replace SM Failed"),
                    (20, "Message Queue is full"),
                    (21, "Invalid Service Type"),
                    (51, "Invalid number of destinations"),
                    (52, "Invalid Distribution List name"),
                    (64, "Invalid Destination Flag (submit_multi)"),
                    (
                        66,
                        "Invalid Submit With Replace request (replace_if_present_flag set)",
                    ),
                    (67, "Invalid esm_class field data"),
                    (68, "Cannot submit to Distribution List"),
                    (69, "submit_sm or submit_multi failed"),
                    (72, "Invalid Source address TON"),
                    (73, "Invalid Source address NPI"),
                    (80, "Invalid Destination address TON"),
                    (81, "Invalid Destination address NPI"),
                    (83, "Invalid system_type field"),
                    (84, "Invalid replace_if_present flag"),
                    (85, "Invalid number of messages"),
                    (88, "Throttling error (ESME has exceeded allowed message limits)"),
                    (97, "Invalid Scheduled Delivery Time"),
                    (98, "Invalid message validity period (Expiry Time)"),
                    (99, "Predefined Message is invalid or not found"),
                    (100, "ESME received Temporary App Error Code"),
                    (101, "ESME received Permanent App Error Code"),
                    (102, "ESME received Reject Message Error Code"),
                    (103, "query_sm request failed"),
                    (192, "Error in the optional part of the PDU body"),
                    (193, "Optional Parameter not allowed"),
                    (194, "Invalid Parameter Length"),
                    (195, "Expected Optional Parameter missing"),
                    (196, "Invalid Optional Parameter Value"),
                    (254, "Delivery Failure (used data_sm_resp)"),
                    (255, "Unknown Error"),
                ],
                null=True,
                verbose_name="command status",
            ),
        ),
    ]



import logging
import select
import socket

from typing import Dict

import smpplib
import smpplib.client
import smpplib.consts
import smpplib.exceptions
import smpplib.gsm

from django.utils import timezone
from rapidsms.models import Backend
from smpplib.command import Command, DeliverSM, SubmitSMResp

from smpp_gateway.models import MOMessage, MTMessage, MTMessageStatus
from smpp_gateway.monitoring import HealthchecksIoWorker
from smpp_gateway.queries import get_mt_messages_to_send, pg_listen, pg_notify
from smpp_gateway.utils import decoded_params, set_exit_signals

logger = logging.getLogger(__name__)


class PgSmppSequenceGenerator(smpplib.client.SimpleSequenceGenerator):
    """
    smpplib sequence generator that uses a Postgres sequence to persist
    sequence numbers across restarts and client instances. See:
    https://www.postgresql.org/docs/10/sql-createsequence.html
    """

    def __init__(self, conn, backend_name):
        self.conn = conn
        self.sequence_name = f"smpp_gateway_sequence_{backend_name}"
        with conn.cursor() as curs:
            curs.execute(
                f"""
                CREATE SEQUENCE IF NOT EXISTS {self.sequence_name}
                    MINVALUE {self.MIN_SEQUENCE}
                    MAXVALUE {self.MAX_SEQUENCE}
                    CYCLE
                """
            )

    def _fetchone(self, query):
        with self.conn.cursor() as curs:
            curs.execute(query)
            return curs.fetchone()[0]

    @property
    def sequence(self):
        "Current (last) value of the sequence."
        return self._fetchone(f"SELECT last_value from {self.sequence_name}")

    def next_sequence(self):
        "Increments and returns the next value of the sequence."
        return self._fetchone(f"SELECT nextval('{self.sequence_name}')")


class PgSmppClient(smpplib.client.Client):
    """
    Thread-safe smpplib Client that waits for messages to send via
    Postgres' LISTEN statement. Loosely adapted from:
    https://stackoverflow.com/a/51105047/166053 and
    https://gist.github.com/pkese/2790749
    """

    def __init__(
        self,
        notify_mo_channel: str,
        backend: Backend,
        hc_worker: HealthchecksIoWorker,
        submit_sm_params: Dict,
        *args,
        **kwargs,
    ):
        self.exit_signal_received = set_exit_signals()
        self.notify_mo_channel = notify_mo_channel
        self.backend = backend
        self.hc_worker = hc_worker
        self.submit_sm_params = submit_sm_params
        super().__init__(*args, **kwargs)
        self._pg_conn = pg_listen(self.backend.name)

    # ############### Handlers ################

    def message_received_handler(self, pdu: DeliverSM):
        """Called by smpplib base Client."""
        mo_params = decoded_params(pdu)
        if mo_params.get("receipted_message_id"):
            self._save_delivery_receipt(pdu, mo_params)
        else:
            self._create_mo_message(pdu, mo_params)

    def _save_delivery_receipt(self, pdu: DeliverSM, params):
        """We received an update that an outbound message was delivered.
        Mark it as delivered.
        """
        count = MTMessageStatus.objects.filter(
            backend=self.backend,
            message_id=params["receipted_message_id"],
        ).update(
            modify_time=timezone.now(),
            delivery_report=pdu.short_message,
        )
        if count == 0:
            logger.warning(
                f"Found no MTMessageStatus for backend={self.backend}, "
                f"message_id={params['receipted_message_id']}. "
                f"delivery_report={pdu.short_message}"
            )
        count = MTMessage.objects.filter(
            backend=self.backend,
            mtmessagestatus__message_id=params["receipted_message_id"],
        ).update(
            modify_time=timezone.now(),
            status=MTMessage.Status.DELIVERED,
        )

    def _create_mo_message(self, pdu: DeliverSM, params):
        """We received a message. Insert into DB and notify the
        listen_mo_messages process.
        """
        now = timezone.now()
        MOMessage.objects.create(
            create_time=now,
            modify_time=now,
            backend=self.backend,
            short_message=pdu.short_message or b"",
            params=params,
            status=MOMessage.Status.NEW,
        )
        pg_notify(self.notify_mo_channel)

    def message_sent_handler(self, pdu: SubmitSMResp):
        """Called by smpplib base Client."""
        params = decoded_params(pdu)
        count = MTMessageStatus.objects.filter(
            backend=self.backend,
            sequence_number=pdu.sequence,
        ).update(
            modify_time=timezone.now(),
            command_status=pdu.status,
            message_id=params["message_id"] or "",
        )
        if count == 0:
            logger.warning(
                f"Found no MTMessageStatus for {self.backend}, {pdu.sequence}. "
                f"status={pdu.status}, message_id={params['message_id']}"
            )

    def error_pdu_handler(self, pdu: Command):
        """Called by smpplib base Client when incoming PDU has status set to
        anything other than OK.
        """
        if pdu.command == "submit_sm_resp":
            # update MTMessageStatus record with the error
            self.message_sent_handler(pdu)
        logger.warning(
            "({}) {}: {}".format(
                pdu.status,
                pdu.command,
                smpplib.consts.DESCRIPTIONS.get(pdu.status, "Unknown status"),
            ),
        )

    # ############### Listen for and send MT Messages ################

    def receive_pg_notifies(self):
        self._pg_conn.poll()
        while self._pg_conn.notifies:
            notify = self._pg_conn.notifies.pop()
            logger.info(f"Got NOTIFY:{notify}")
            self.send_mt_messages()

    def send_mt_messages(self):
        smses = get_mt_messages_to_send(limit=100, backend=self.backend)
        submit_sm_resps = []
        for sms in smses:
            params = {**self.submit_sm_params, **sms["params"]}
            pdus = self.split_and_send_message(sms["short_message"], **params)
            # Create placeholder MTMessageStatus objects in the DB, which
            # the message_sent handler will later update with the actual command_status
            # and message_id (and eventually maybe a delivery report).
            now = timezone.now()
            submit_sm_resps.extend(
                [
                    MTMessageStatus(
                        create_time=now,
                        modify_time=now,
                        mt_message_id=sms["id"],
                        backend=self.backend,
                        sequence_number=pdu.sequence,
                    )
                    for pdu in pdus
                ]
            )
        pks = [sms["id"] for sms in smses]
        MTMessage.objects.filter(pk__in=pks).update(
            status=MTMessage.Status.SENT,
            modify_time=timezone.now(),
        )
        MTMessageStatus.objects.bulk_create(submit_sm_resps)

    def split_and_send_message(self, message, **kwargs):
        """
        Splits and sends the given message, returning the underlying PDUs.
        The "source_addr" and "destination_addr" keyword arguments are required
        by python-smpplib.
        """
        # Two parts, UCS2, SMS with UDH
        parts, data_coding, esm_class = smpplib.gsm.make_parts(message)
        return [
            self.send_message(
                short_message=short_message,
                data_coding=data_coding,
                esm_class=esm_class,
                **kwargs,
            )
            for short_message in parts
        ]

    # ############### Main loop ################

    def listen(self, ignore_error_codes=None, auto_send_enquire_link=True):
        self.logger.info("Entering main listen loop")
        # Look for and send up to 100 messages on start up
        self.send_mt_messages()
        while True:
            # When either main socket has data or _pg_conn has data, select.select will return
            rlist, _, _ = select.select(
                [self._socket, self._pg_conn], [], [], self.timeout
            )
            if not rlist and auto_send_enquire_link:
                self.logger.debug("Socket timeout, listening again")
                pdu = smpplib.smpp.make_pdu("enquire_link", client=self)
                self.send_pdu(pdu)
                # Look for and send up to 100 messages every 5 seconds
                self.send_mt_messages()
                continue
            elif not rlist:
                # backwards-compatible with existing behavior
                raise socket.timeout()
            for ready_socket in rlist:
                if ready_socket is self._socket:
                    self.read_once(ignore_error_codes, auto_send_enquire_link)
                else:
                    self.receive_pg_notifies()
            if self.hc_worker:
                self.hc_worker.success_ping()
            if self.exit_signal_received():
                self.logger.info("Got exit signal, leaving listen loop")
                self.safe_disconnect()
                return

    def safe_disconnect(self):
        if self._socket is not None:
            try:
                self.logger.info("Unbinding...")
                self.unbind()
            except:
                self.logger.exception("Ignoring exception during unbind")
            try:
                # Disconnect writes its own "Disconnecting..." message
                self.disconnect()
            except:
                self.logger.exception("Ignoring exception during disconnect")



import logging
import queue
import threading
import time

logger = logging.getLogger(__name__)


class HealthchecksIoWorker:
    """
    Worker class to ping healthchecks.io from a separate thread (avoids holding
    up the main loop in case of connectivity issues).
    """

    SUCCESS = 1
    FAIL = 2

    def __init__(self, uuid=None, ping_key=None, slug=None):
        # Only attempt import if healthchecks.io is enabled
        from healthchecks_io import Client

        self.queue = queue.Queue()
        self.client = Client(ping_key=ping_key)
        if uuid:
            self.ping_kwargs = {"uuid": uuid}
        else:
            self.ping_kwargs = {"slug": slug}
        threading.Thread(target=self._worker_thread, daemon=True).start()

    def _worker_thread(self):
        from healthchecks_io.client.exceptions import HCAPIError

        last_success_time = 0
        while True:
            item = self.queue.get()
            try:
                if item == HealthchecksIoWorker.SUCCESS:
                    # Skip success_ping if less than a minute has passed since the last one
                    # (the highest frequency possible in healthchecks.io / cron syntax).
                    if last_success_time + 60 < time.time():
                        logger.debug(f"Sending success_ping for {self.ping_kwargs}")
                        self.client.success_ping(**self.ping_kwargs)
                        last_success_time = time.time()
                    else:
                        logger.debug(f"Skipping success_ping for {self.ping_kwargs}")
                elif item == HealthchecksIoWorker.FAIL:
                    logger.debug(f"Sending fail_ping for {self.ping_kwargs}")
                    self.client.fail_ping(**self.ping_kwargs)
                else:
                    raise ValueError(f"Unknown task: {item}")
            except HCAPIError:
                logger.warn("Failed to ping healthchecks.io API", exc_info=True)
            self.queue.task_done()

    def success_ping(self):
        self.queue.put(HealthchecksIoWorker.SUCCESS)

    def fail_ping(self):
        self.queue.put(HealthchecksIoWorker.FAIL)

    def join(self):
        # Block until all tasks are done.
        self.queue.join()



# Generated by Django 2.2.24 on 2021-12-06 15:55

import django.db.models.deletion

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rapidsms", "0004_auto_20150801_2138"),
        ("smpp_gateway", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="momessage",
            options={"verbose_name": "mobile-originated message"},
        ),
        migrations.AlterModelOptions(
            name="mtmessage",
            options={"verbose_name": "mobile-terminated message"},
        ),
        migrations.CreateModel(
            name="MTMessageStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("create_time", models.DateTimeField()),
                ("modify_time", models.DateTimeField()),
                ("sequence_number", models.IntegerField()),
                ("command_status", models.IntegerField(null=True)),
                (
                    "message_id",
                    models.CharField(blank=True, default="", max_length=255),
                ),
                ("delivery_report", models.BinaryField(null=True)),
                (
                    "backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="rapidsms.Backend",
                    ),
                ),
                (
                    "mt_message",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="smpp_gateway.MTMessage",
                    ),
                ),
            ],
            options={
                "verbose_name": "mobile-terminated message status",
            },
        ),
        migrations.AddConstraint(
            model_name="mtmessagestatus",
            constraint=models.UniqueConstraint(
                fields=("backend", "sequence_number"), name="unique_seq_num"
            ),
        ),
    ]



import logging
import select

import psycopg2.extensions

from django.db.models import QuerySet
from rapidsms.router import lookup_connections, receive

from smpp_gateway.models import MOMessage
from smpp_gateway.queries import get_mo_messages_to_process, pg_listen
from smpp_gateway.utils import set_exit_signals

logger = logging.getLogger(__name__)


def handle_mo_messages(smses: QuerySet[MOMessage]):
    received_smses = []
    try:
        for sms in smses:
            connection = lookup_connections(
                backend=sms.backend, identities=[sms.params["source_addr"]]
            )[0]
            fields = {
                "to_addr": sms.params["destination_addr"],
                "from_addr": sms.params["source_addr"],
            }
            receive(sms.decoded_short_message, connection, fields=fields)
            received_smses.append(sms)
    finally:
        if received_smses:
            MOMessage.objects.filter(pk__in=[sms.pk for sms in received_smses]).update(
                status=MOMessage.Status.DONE
            )


def listen_mo_messages(channel: str):
    """Batch process any queued incoming messages, then listen to be notified
    of new arrivals.
    """
    exit_signal_received = set_exit_signals()
    smses = get_mo_messages_to_process(limit=100)
    while smses:
        handle_mo_messages(smses)
        # If an exit was triggered, do so before retrieving more messages to process...
        if exit_signal_received():
            logger.info("Received exit signal, leaving processing loop...")
            return
        smses = get_mo_messages_to_process(limit=100)

    pg_conn = pg_listen(channel)

    while True:
        if select.select([pg_conn], [], [], 5) == ([], [], []):
            logger.debug(f"{channel} .")
        else:
            pg_conn.poll()
            while pg_conn.notifies:
                notify = pg_conn.notifies.pop()  # type: psycopg2.extensions.Notify
                logger.info(f"Got NOTIFY:{notify}")
                smses = get_mo_messages_to_process(limit=1)
                handle_mo_messages(smses)
        if exit_signal_received():
            logger.info("Received exit signal, leaving listen loop...")
            return



import json
import logging

from typing import Dict, Optional

from django.db import connection as db_conn
from rapidsms.models import Backend

from smpp_gateway.client import PgSmppClient, PgSmppSequenceGenerator
from smpp_gateway.monitoring import HealthchecksIoWorker

logger = logging.getLogger(__name__)


def get_smpplib_client(
    host: str,
    port: int,
    notify_mo_channel: str,
    backend: Backend,
    submit_sm_params: Dict,
    hc_check_uuid: str,
    hc_ping_key: str,
    hc_check_slug: str,
) -> PgSmppClient:
    sequence_generator = PgSmppSequenceGenerator(db_conn, backend.name)
    if hc_check_uuid:
        hc_worker = HealthchecksIoWorker(uuid=hc_check_uuid)
    elif hc_ping_key and hc_check_slug:
        hc_worker = HealthchecksIoWorker(ping_key=hc_ping_key, slug=hc_check_slug)
    else:
        hc_worker = None
    client = PgSmppClient(
        notify_mo_channel,
        backend,
        hc_worker,
        submit_sm_params,
        host,
        port,
        allow_unknown_opt_params=True,
        sequence_generator=sequence_generator,
    )
    return client


def smpplib_main_loop(
    client: PgSmppClient,
    system_id: str,
    password: str,
    interface_version: Optional[str],
    system_type: Optional[str],
):
    client.connect()
    client.bind_transceiver(
        system_id=system_id,
        password=password,
        interface_version=interface_version,
        system_type=system_type,
    )
    client.listen()


def start_smpp_client(options):
    backend, _ = Backend.objects.get_or_create(name=options["backend_name"])
    client = get_smpplib_client(
        options["host"],
        options["port"],
        options["notify_mo_channel"],
        backend,
        json.loads(options["submit_sm_params"]),
        options["hc_check_uuid"],
        options["hc_ping_key"],
        options["hc_check_slug"],
    )
    smpplib_main_loop(
        client,
        options["system_id"],
        options["password"],
        options["interface_version"],
        options["system_type"],
    )



from django.core.management.base import BaseCommand

from smpp_gateway.subscribers import listen_mo_messages


class Command(BaseCommand):
    help = "Receive inbound MO messages and pass them to RapidSMS."

    def add_arguments(self, parser):
        parser.add_argument("--channel", default="new_mo_msg")

    def handle(self, *args, **options):
        listen_mo_messages(channel=options["channel"])



from django.contrib import admin
from smpplib.consts import DESCRIPTIONS

from smpp_gateway.models import MOMessage, MTMessage, MTMessageStatus


@admin.register(MOMessage)
class MOMessageAdmin(admin.ModelAdmin):
    list_display = (
        "decoded_short_message",
        "backend",
        "status",
        "create_time",
    )
    list_filter = ("status", "backend")
    ordering = ("-create_time",)


class MTMessageStatusInline(admin.TabularInline):
    model = MTMessageStatus
    readonly_fields = (
        "create_time",
        "modify_time",
        "backend",
        "sequence_number",
        "command_status",
        "message_id",
        "delivery_report_as_bytes",
    )
    can_delete = False
    extra = 0


class MTMessageCommandStatusListFilter(admin.SimpleListFilter):
    title = "command status"
    parameter_name = "command_status"
    path_to_parameter = "mtmessagestatus__command_status"

    def lookups(self, request, model_admin):
        """
        Only show the lookups for command_status values that exist in the database.
        """
        statuses = MTMessageStatus.objects.values_list(
            "command_status", flat=True
        ).distinct()
        return [(status, DESCRIPTIONS.get(status, "")) for status in statuses]

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset by command_status.
        """
        if self.value():
            return queryset.filter(**{self.path_to_parameter: self.value()})
        return queryset


@admin.register(MTMessage)
class MTMessageAdmin(admin.ModelAdmin):
    list_display = (
        "short_message",
        "backend",
        "status",
        "create_time",
    )
    list_filter = (
        "status",
        "backend",
        MTMessageCommandStatusListFilter,
    )
    search_fields = ("mtmessagestatus__sequence_number",)
    ordering = ("-create_time",)
    inlines = (MTMessageStatusInline,)


class MTMessageStatusCommandStatusListFilter(MTMessageCommandStatusListFilter):
    path_to_parameter = "command_status"


@admin.register(MTMessageStatus)
class MTMessageStatusAdmin(admin.ModelAdmin):
    list_display = (
        "sequence_number",
        "backend",
        "message_id",
        "command_status",
        "create_time",
    )
    list_filter = (
        "backend",
        MTMessageStatusCommandStatusListFilter,
    )
    search_fields = ("sequence_number",)
    ordering = ("-create_time",)



import pytest

from smpplib import consts as smpplib_consts
from smpplib.command import DeliverSM, SubmitSMResp

from smpp_gateway.models import MOMessage, MTMessage
from smpp_gateway.queries import pg_listen
from smpp_gateway.smpp import get_smpplib_client
from tests.factories import BackendFactory, MTMessageFactory, MTMessageStatusFactory


@pytest.mark.django_db(transaction=True)
class TestMessageReceivedHandler(object):
    def test_received_mo_message(self):
        """When a message is received that is not a receipt
        acknowledgement, persist it and notify the MO listener.
        """
        backend = BackendFactory()
        client = get_smpplib_client(
            "127.0.0.1",
            8000,
            "notify_mo_channel",
            backend,
            {},  # submit_sm_params
            "",  # hc_check_uuid
            "",  # hc_ping_key
            "",  # hc_check_slug
        )

        pdu = DeliverSM("deliver_sm")
        pdu.short_message = b"this is a short message"
        pdu.source_addr = "+46166371876"

        listen_conn = pg_listen("notify_mo_channel")

        client.message_received_handler(pdu)

        # should notify the MO listener
        listen_conn.poll()
        assert len(listen_conn.notifies) == 1

        # there should be a message for the MO listener to process
        msg = MOMessage.objects.get()
        assert msg.backend == backend
        assert msg.short_message.tobytes() == pdu.short_message
        assert msg.params["source_addr"] == "+46166371876"
        assert msg.status == MOMessage.Status.NEW

    def test_received_message_receipt(self):
        """When a message is received that is a receipt acknowledgement,
        persist the delivery report on the appropriate outbound messages.
        """
        backend = BackendFactory()
        client = get_smpplib_client(
            "127.0.0.1",
            8000,
            "notify_mo_channel",
            backend,
            {},  # submit_sm_params
            "",  # hc_check_uuid
            "",  # hc_ping_key
            "",  # hc_check_slug
        )
        outbound_msg = MTMessageFactory(status=MTMessage.Status.SENT, backend=backend)
        outbound_msg_status = MTMessageStatusFactory(
            mt_message=outbound_msg, message_id="abcdefg"
        )

        pdu = DeliverSM("deliver_sm")
        pdu.short_message = b"this is a delivery receipt"
        pdu.source_addr = "+46166371876"
        pdu.receipted_message_id = "abcdefg"

        listen_conn = pg_listen("notify_mo_channel")

        client.message_received_handler(pdu)

        # should *not* notify the MO listener
        listen_conn.poll()
        assert len(listen_conn.notifies) == 0

        outbound_msg.refresh_from_db()
        outbound_msg_status.refresh_from_db()

        assert outbound_msg.status == MTMessage.Status.DELIVERED
        assert (
            outbound_msg_status.delivery_report.tobytes()
            == b"this is a delivery receipt"
        )

    def test_received_null_short_message(self):
        """When a message is received with a null short_message, persist it
        and notify the MO listener."""
        backend = BackendFactory()
        client = get_smpplib_client(
            "127.0.0.1",
            8000,
            "notify_mo_channel",
            backend,
            {},  # submit_sm_params
            "",  # hc_check_uuid
            "",  # hc_ping_key
            "",  # hc_check_slug
        )

        pdu = DeliverSM("deliver_sm")
        pdu.short_message = None
        pdu.source_addr = "+46166371876"

        listen_conn = pg_listen("notify_mo_channel")

        client.message_received_handler(pdu)

        # should notify the MO listener
        listen_conn.poll()
        assert len(listen_conn.notifies) == 1

        # there should be a message for the MO listener to process
        msg = MOMessage.objects.get()
        assert msg.backend == backend
        assert msg.short_message == b""
        assert msg.params["source_addr"] == "+46166371876"
        assert msg.status == MOMessage.Status.NEW


@pytest.mark.django_db(transaction=True)
def test_message_sent_handler():
    """The associated MTMessageStatus should be updated with the submission
    status and message_id.
    """
    backend = BackendFactory()
    client = get_smpplib_client(
        "127.0.0.1",
        8000,
        "notify_mo_channel",
        backend,
        {},  # submit_sm_params
        "",  # hc_check_uuid
        "",  # hc_ping_key
        "",  # hc_check_slug
    )
    outbound_msg = MTMessageFactory(status=MTMessage.Status.SENT, backend=backend)
    outbound_msg_status = MTMessageStatusFactory(mt_message=outbound_msg)

    pdu = SubmitSMResp("submit_sm_resp")
    pdu.sequence = outbound_msg_status.sequence_number
    pdu.status = smpplib_consts.SMPP_ESME_RSUBMITFAIL
    pdu.message_id = "qwerty"

    client.message_sent_handler(pdu)

    outbound_msg_status.refresh_from_db()

    assert outbound_msg_status.command_status == smpplib_consts.SMPP_ESME_RSUBMITFAIL
    assert outbound_msg_status.message_id == "qwerty"



import os

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": os.getenv("TEST_DB_NAME", "test_database"),
        "USER": os.getenv("TEST_DB_USER", "test_user"),
        "PASSWORD": os.getenv("TEST_DB_PASSWORD", "test_password"),
        "HOST": os.getenv("TEST_DB_HOST", "127.0.0.1"),
        "PORT": os.getenv("TEST_DB_PORT", "5432"),
    }
}

INSTALLED_APPS = [
    "rapidsms",
    "smpp_gateway",
]



import os

from django.core.management.base import BaseCommand

from smpp_gateway.smpp import start_smpp_client


def default_port():
    port = os.environ.get("SMPPLIB_PORT", 2775)
    return int(port) if port else None


class Command(BaseCommand):
    help = "Start an SMPP client instance."

    def add_arguments(self, parser):
        parser.add_argument(
            "backend_name",
            help="RapidSMS backend name. Will be created if it doesn't exist.",
        )
        parser.add_argument(
            "--notify-mo-channel",
            help="Name of Postgres channel to NOTIFY for each incoming message.",
            default="new_mo_msg",
        )
        parser.add_argument(
            "--host",
            default=os.environ.get("SMPPLIB_HOST"),
        )
        parser.add_argument(
            "--port",
            default=default_port(),
        )
        parser.add_argument(
            "--system-id",
            default=os.environ.get("SMPPLIB_SYSTEM_ID"),
        )
        parser.add_argument(
            "--password",
            default=os.environ.get("SMPPLIB_PASSWORD"),
        )
        parser.add_argument(
            "--system-type",
            default=os.environ.get("SMPPLIB_SYSTEM_TYPE"),
        )
        parser.add_argument(
            "--interface-version",
            default=os.environ.get("SMPPLIB_INTERFACE_VERSION"),
        )
        parser.add_argument(
            "--submit-sm-params",
            default=os.environ.get("SMPPLIB_SUBMIT_SM_PARAMS", r"{}"),
        )
        parser.add_argument(
            "--database-url",
            default=os.environ.get("DATABASE_URL"),
        )
        parser.add_argument("--log-file")
        parser.add_argument(
            "--send-bulksms",
            type=int,
            help="Sends the specified number of SMSes in bulk on startup. "
            "Not intended for use with a real MNO.",
        )
        parser.add_argument(
            "--hc-check-uuid",
            default=os.environ.get("HEALTHCHECKS_IO_CHECK_UUID"),
            help="Pings healthchecks.io with the specified check UUID. "
            "If set, --hc-ping-key and --hc-check-slug will be ignored.",
        )
        parser.add_argument(
            "--hc-ping-key",
            default=os.environ.get("HEALTHCHECKS_IO_PING_KEY"),
            help="Pings healthchecks.io with the specified ping key and check slug. "
            "If set, --hc-check-slug must also be set.",
        )
        parser.add_argument(
            "--hc-check-slug",
            default=os.environ.get("HEALTHCHECKS_IO_CHECK_SLUG"),
            help="Pings healthchecks.io with the specified ping key and check slug. "
            "If set, --hc-ping-key must also be set.",
        )

    def handle(self, *args, **options):
        start_smpp_client(options)



import base64
import itertools
import logging
import signal
import string

from typing import Any, Dict

import smpplib

ASCII_PRINTABLE_BYTES = {ord(c) for c in string.printable}

logger = logging.getLogger(__name__)


def set_exit_signals(signals=[signal.SIGINT, signal.SIGTERM]):
    """
    Helper method for handling graceful exit signals such as SIGINT and SIGTERM.

    Use like so:

    exit_signal_received = set_exit_signals()
    while True:
        get_and_process_some_data()
        if exit_signal_received():
            logger.info("Received exit signal, leaving loop...")
            return
    """
    _received_exit_signal = False

    def set_received_exit_signal(sig_num, stack_frame):
        nonlocal _received_exit_signal
        logger.info(f"Got signal {sig_num}, setting exit_signal_received = True")
        _received_exit_signal = True

    for sig in signals:
        signal.signal(sig, set_received_exit_signal)

    return lambda: _received_exit_signal


def grouper(iterable, n):
    """
    Collect data into fixed-length chunks or blocks. Handy for grouping iterables of
    objects into batches suitable for bulk_create().
    """
    # In case iterable is not an iterator (but is, e.g., a list), islice()
    # will return the first group indefinitely, so ensure iterable is an
    # iterator first.
    iterable = iter(iterable)
    while True:
        group = list(itertools.islice(iterable, n))
        if not group:
            break
        yield group


def maybe_decode(value):
    if isinstance(value, bytes):
        if all(b in ASCII_PRINTABLE_BYTES for b in value):
            return value.decode("ascii")
        else:
            return base64.b64encode(value).decode("utf-8")
    return value


def decoded_params(pdu: smpplib.command.Command) -> Dict[str, Any]:
    return {key: maybe_decode(getattr(pdu, key)) for key in pdu.params.keys()}



# Generated by Django 2.2.24 on 2021-12-03 04:06

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("rapidsms", "0004_auto_20150801_2138"),
    ]

    operations = [
        migrations.CreateModel(
            name="MTMessage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("create_time", models.DateTimeField()),
                ("modify_time", models.DateTimeField()),
                ("short_message", models.TextField()),
                ("params", django.contrib.postgres.fields.jsonb.JSONField()),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("new", "New"),
                            ("sending", "Sending"),
                            ("sent", "Sent"),
                            ("delivered", "Delivered"),
                            ("error", "Error"),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    "backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="rapidsms.Backend",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MOMessage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("create_time", models.DateTimeField()),
                ("modify_time", models.DateTimeField()),
                ("short_message", models.BinaryField()),
                ("params", django.contrib.postgres.fields.jsonb.JSONField()),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("new", "New"),
                            ("processing", "Processing"),
                            ("done", "Done"),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    "backend",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="rapidsms.Backend",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.0.5 on 2022-06-30 11:40

import django.db.models.deletion

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rapidsms", "0004_auto_20150801_2138"),
        ("smpp_gateway", "0003_big_pks_and_help_texts"),
    ]

    operations = [
        migrations.AlterField(
            model_name="momessage",
            name="backend",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="rapidsms.backend",
                verbose_name="backend",
            ),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="create_time",
            field=models.DateTimeField(verbose_name="create time"),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="modify_time",
            field=models.DateTimeField(verbose_name="modify time"),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="params",
            field=models.JSONField(verbose_name="params"),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="short_message",
            field=models.BinaryField(verbose_name="short message"),
        ),
        migrations.AlterField(
            model_name="momessage",
            name="status",
            field=models.CharField(
                choices=[
                    ("new", "New"),
                    ("processing", "Processing"),
                    ("done", "Done"),
                ],
                max_length=32,
                verbose_name="status",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="backend",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="rapidsms.backend",
                verbose_name="backend",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="create_time",
            field=models.DateTimeField(verbose_name="create time"),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="modify_time",
            field=models.DateTimeField(verbose_name="modify time"),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="params",
            field=models.JSONField(verbose_name="params"),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="short_message",
            field=models.TextField(verbose_name="short message"),
        ),
        migrations.AlterField(
            model_name="mtmessage",
            name="status",
            field=models.CharField(
                choices=[
                    ("new", "New"),
                    ("sending", "Sending"),
                    ("sent", "Sent"),
                    ("delivered", "Delivered"),
                    ("error", "Error"),
                ],
                max_length=32,
                verbose_name="status",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="backend",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="rapidsms.backend",
                verbose_name="backend",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="command_status",
            field=models.IntegerField(null=True, verbose_name="command status"),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="create_time",
            field=models.DateTimeField(verbose_name="create time"),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="delivery_report",
            field=models.BinaryField(
                help_text="The delivery_report we receive from the MC (via its deliver_sm).",
                null=True,
                verbose_name="delivery report",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="message_id",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The message_id we receive from the MC (via its submit_sm_resp).",
                max_length=255,
                verbose_name="message id",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="modify_time",
            field=models.DateTimeField(verbose_name="modify time"),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="mt_message",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="smpp_gateway.mtmessage",
                verbose_name="mt message",
            ),
        ),
        migrations.AlterField(
            model_name="mtmessagestatus",
            name="sequence_number",
            field=models.IntegerField(
                help_text="The initial sequence_number established by us when the message was sent.",
                verbose_name="sequence number",
            ),
        ),
    ]



from django.db import connection, models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rapidsms.models import Backend
from smpplib.consts import DESCRIPTIONS


class AbstractTimestampModel(models.Model):
    create_time = models.DateTimeField(_("create time"))
    modify_time = models.DateTimeField(_("modify time"))

    class Meta:
        abstract = True


class MOMessage(AbstractTimestampModel, models.Model):
    """Mobile-originated, or inbound, message."""

    class Status(models.TextChoices):
        NEW = "new", _("New")
        PROCESSING = "processing", _("Processing")
        DONE = "done", _("Done")

    backend = models.ForeignKey(
        Backend, on_delete=models.PROTECT, verbose_name=_("backend")
    )
    # Save the raw bytes, in case they're needed later
    short_message = models.BinaryField(_("short message"))
    params = models.JSONField(_("params"))
    status = models.CharField(_("status"), max_length=32, choices=Status.choices)

    @cached_property
    def decoded_short_message(self) -> str:
        data_coding = self.params.get("data_coding", 0)
        short_message = self.short_message.tobytes()  # type: bytes
        if data_coding == 0:
            return short_message.decode("ascii")
        if data_coding == 8:
            return short_message.decode("utf-16-be")
        raise ValueError(
            f"Unsupported data_coding {data_coding}. Short message: {short_message}"
        )

    def __str__(self):
        return f"{self.params['short_message']} ({self.id})"

    class Meta:
        verbose_name = _("mobile-originated message")


class MTMessage(AbstractTimestampModel, models.Model):
    """Mobile-terminated, or outbound, message."""

    class Status(models.TextChoices):
        NEW = "new", _("New")
        SENDING = "sending", _("Sending")
        SENT = "sent", _("Sent")
        DELIVERED = "delivered", _("Delivered")
        ERROR = "error", _("Error")

    backend = models.ForeignKey(
        Backend, on_delete=models.PROTECT, verbose_name=_("backend")
    )
    # SMPP client will decide how to encode it
    short_message = models.TextField(_("short message"))
    params = models.JSONField(_("params"))
    status = models.CharField(_("status"), max_length=32, choices=Status.choices)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.status == MTMessage.Status.NEW:
            with connection.cursor() as curs:
                curs.execute(f"NOTIFY {self.backend.name}")

    def __str__(self):
        return f"{self.short_message} ({self.id})"

    class Meta:
        verbose_name = _("mobile-terminated message")


class MTMessageStatus(AbstractTimestampModel, models.Model):
    """Metadata and status information about outbound messages."""

    COMMAND_STATUS_CHOICES = list(DESCRIPTIONS.items())
    mt_message = models.ForeignKey(
        MTMessage, on_delete=models.CASCADE, verbose_name=_("mt message")
    )
    backend = models.ForeignKey(
        Backend, on_delete=models.PROTECT, verbose_name=_("backend")
    )
    sequence_number = models.IntegerField(
        _("sequence number"),
        help_text=_(
            "The initial sequence_number established by us when the message was sent."
        ),
    )
    command_status = models.IntegerField(
        _("command status"), null=True, choices=COMMAND_STATUS_CHOICES
    )
    message_id = models.CharField(
        _("message id"),
        max_length=255,
        blank=True,
        default="",
        help_text=_("The message_id we receive from the MC (via its submit_sm_resp)."),
    )
    delivery_report = models.BinaryField(
        _("delivery report"),
        null=True,
        help_text=_("The delivery_report we receive from the MC (via its deliver_sm)."),
    )

    @cached_property
    def delivery_report_as_bytes(self) -> bytes:
        return self.delivery_report.tobytes()

    def __str__(self):
        return f"{self.backend} ({self.sequence_number})"

    class Meta:
        verbose_name = _("mobile-terminated message status")
        verbose_name_plural = _("mobile-terminated message statuses")
        constraints = [
            models.UniqueConstraint(
                fields=["backend", "sequence_number"], name="unique_seq_num"
            )
        ]



from unittest import mock

import pytest

from smpp_gateway.models import MOMessage
from smpp_gateway.subscribers import handle_mo_messages
from tests.factories import MOMessageFactory


@pytest.mark.django_db
class TestHandleMOMessages(object):
    def receive_was_called(self, mock_receive: mock.Mock, message: MOMessage):
        return len(
            [
                call
                for call in mock_receive.call_args_list
                if call.args[0] == message.decoded_short_message
            ]
        )

    def test_mark_received_messages_done(self):
        """Handle each incoming message by receving them into RapidSMS and
        marking them as done.
        """
        MOMessageFactory.create_batch(5)
        messages = MOMessage.objects.all()

        with mock.patch("smpp_gateway.subscribers.receive") as mock_receive:
            handle_mo_messages(messages)

        for message in messages:
            message.refresh_from_db()
            assert message.status == MOMessage.Status.DONE
            assert self.receive_was_called(mock_receive, message)

    def test_error_in_middle_of_patch(self):
        """If receiving one message raises an exception, ensure all received
        messages are still marked as done to avoid duplicate receiving.

        Unreceived messages should be left untouched to be retried later.
        """
        MOMessageFactory.create_batch(5)
        messages = MOMessage.objects.all()

        with mock.patch("smpp_gateway.subscribers.receive") as mock_receive:
            mock_receive.side_effect = [None, None, Exception("boom"), None, None]
            with pytest.raises(Exception, match=r"boom"):
                handle_mo_messages(messages)

        for received_message in messages[:2]:
            # Messages 1, 2 should be received and marked as done
            received_message.refresh_from_db()
            assert received_message.status == MOMessage.Status.DONE
            assert self.receive_was_called(mock_receive, received_message)

        # Message 3 raised an exception, should not be marked done so it
        # will be retried later
        failed_message = messages[2]
        failed_message.refresh_from_db()
        assert failed_message.status == MOMessage.Status.NEW
        assert self.receive_was_called(mock_receive, failed_message)

        for other_message in messages[3:]:
            # Messages 4, 5 should remain new to be received later
            other_message.refresh_from_db()
            assert other_message.status == MOMessage.Status.NEW
            assert not self.receive_was_called(mock_receive, other_message)
